"""
This module contains the API classes for :py:mod:`kinisi`.
It is anticipated that this is where the majority of interaction with the package will occur.
This module includes the :py:class:`~kinisi.analyze.DiffAnalyzer` class for diffusion analysis, which is compatible with both VASP Xdatcar output files and any MD trajectory that the :py:mod:`MDAnalysis` package can handle.
"""

# Copyright (c) Andrew R. McCluskey and Benjamin J. Morgan
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

import numpy as np
from uravu.distribution import Distribution
from kinisi import diffusion
from kinisi.parser import MDAnalysisParser, PymatgenParser


class MSDAnalyzer:
    """
    The :py:class:`kinisi.analyze.MSDAnalyzer` class evaluates the MSD of atoms in a material.
    This is achieved through the application of a block bootstrapping methodology to obtain the most statistically accurate values for mean squared displacement and the associated uncertainty.

    Attributes:
        dt (:py:attr:`array_like`):  Timestep values.
        distributions (:py:attr:`list` or :py:class:`Distribution`): The distributions describing the MSD at each timestep.
        msd_observed (:py:attr:`array_like`): The sample mean-squared displacements, found from the arithmetic average of the observations.
        msd_sampled (:py:attr:`array_like`): The population mean-squared displacements, found from the bootstrap resampling of the observations.
        msd_sampled_err (:py:attr:`array_like`): The two-dimensional uncertainties, at the given confidence interval, found from the bootstrap resampling of the observations.
        _diff (:py:class:`kinisi.diffusion.MSDBootstrap`): The :py:mod:`kinisi` bootstrap and diffusion object.

    Args:
        trajectory (:py:attr:`str` or :py:attr:`list` of :py:attr:`str` or :py:attr:`list` of :py:class:`pymatgen.core.structure.Structure` or :py:class:`MDAnalysis.core.Universe`): The file path(s) that should be read by either the :py:class:`pymatgen.io.vasp.Xdatcar` or :py:class:`MDAnalysis.core.universe.Universe` classes, a :py:attr:`list` of :py:class:`pymatgen.core.structure.Structure` objects ordered in sequence of run, or an :py:class:`MDAnalysis.core.universe.Universe` object.
        parser_params (:py:attr:`dict`): The parameters for the :py:mod:`kinisi.parser` object, which is either :py:class:`kinisi.parser.PymatgenParser` or :py:class:`kinisi.parser.MDAnalysisParser` depending on the input file format. See the appropriate documention for more guidance on this dictionary.
        bootstrap_params (:py:attr:`dict`, optional): The parameters for the :py:class:`kinisi.diffusion.MSDBootstrap` object. See the appropriate documentation for more guidance on this. Default is the default bootstrap parameters.
        dtype (:py:attr:`str`, optional): The file format of the :py:attr:`trajectory`, for a trajectory file path to be read by :py:class:`pymatgen.io.vasp.Xdatcar` this should be :py:attr:`'Xdatcar'`, for a :py:attr:`list` of :py:class:`pymatgen.core.structure.Structure` objects this should be :py:attr:`'structures'`, for a trajectory file path to be read by :py:mod:`MDAnalysis` this should be the appropriate format to be passed to the :py:class:`MDAnalysis.core.universe.Universe`, and for a n :py:class:`MDAnalysis.core.universe.Universe` object this should be :py:attr:`'universe'`. Defaults to :py:attr:`'Xdatcar'`.
    """
    def __init__(self, trajectory, parser_params, bootstrap_params=None, dtype='Xdatcar'):  # pragma: no cover
        if bootstrap_params is None:
            bootstrap_params = {}
        if dtype is 'Xdatcar':
            try:
                from pymatgen.io.vasp import Xdatcar
            except ModuleNotFoundError:
                raise ModuleNotFoundError("To use the Xdatcar file parsing, pymatgen must be installed.")
            if isinstance(trajectory, list):
                trajectory_list = (Xdatcar(f) for f in trajectory)
                structures = _flatten_list([x.structures for x in trajectory_list])
            else:
                xd = Xdatcar(trajectory)
                structures = xd.structures
            u = PymatgenParser(structures, **parser_params)
            self.first_structure = structures[0]
        elif dtype is 'structures':
            u = PymatgenParser(trajectory, **parser_params)
            self.first_structure = structures[0]
        elif dtype == 'universe':
            u = MDAnalysisParser(trajectory, **parser_params)
        else:
            try:
                import MDAnalysis as mda
            except ModuleNotFoundError:
                raise ModuleNotFoundError("To use the MDAnalysis from file parsing, MDAnalysis must be installed.")
            universe = mda.Universe(*trajectory, format=dtype)
            u = MDAnalysisParser(universe, **parser_params)

        dt = u.delta_t
        disp_3d = u.disp_3d

        self._diff = diffusion.MSDBootstrap(dt, disp_3d, **bootstrap_params)

        self.dt = self._diff.dt
        self.msd_sampled = self._diff.msd_sampled
        self.msd_sampled_err = self._diff.msd_sampled_err
        self.msd_observed = self._diff.msd_observed
        self.distributions = self._diff.distributions

    @property
    def msd(self):
        """
        Returns MSD for the input trajectories. Note that this is the bootstrap sampled MSD, not the numerical average from the data.

        Returns:
            :py:attr:`array_like`: MSD values.
        """
        return self.msd_sampled

    @property
    def msd_err(self):
        """
        Returns MSD uncertainty for the input trajectories.

        Returns:
            :py:attr:`array_like`: A lower and upper uncertainty, at a single standard deviation, of the mean squared displacement values.
        """
        return self.msd_sampled_err


class DiffAnalyzer(MSDAnalyzer):
    """
    The :py:class:`kinisi.analyze.DiffAnalyzer` class performs analysis of diffusion relationships in materials.
    This is achieved through the application of a block bootstrapping methodology to obtain the most statistically accurate values for mean squared displacement and the associated uncertainty.
    The time-scale dependence of the MSD is then modeled with a straight line Einstein relationship, and Markov chain Monte Carlo is used to quantify inverse uncertainties for this model.

    Args:
        trajectory (:py:attr:`str` or :py:attr:`list` of :py:attr:`str` or :py:attr:`list` of :py:class:`pymatgen.core.structure.Structure` or :py:class:`MDAnalysis.core.Universe`): The file path(s) that should be read by either the :py:class:`pymatgen.io.vasp.Xdatcar` or :py:class:`MDAnalysis.core.universe.Universe` classes, a :py:attr:`list` of :py:class:`pymatgen.core.structure.Structure` objects ordered in sequence of run, or an :py:class:`MDAnalysis.core.universe.Universe` object.
        parser_params (:py:attr:`dict`): The parameters for the :py:mod:`kinisi.parser` object, which is either :py:class:`kinisi.parser.PymatgenParser` or :py:class:`kinisi.parser.MDAnalysisParser` depending on the input file format. See the appropriate documention for more guidance on this dictionary.
        bootstrap_params (:py:attr:`dict`, optional): The parameters for the :py:class:`kinisi.diffusion.MSDBootstrap` object. See the appropriate documentation for more guidance on this. Default is the default bootstrap parameters.
        dtype (:py:attr:`str`, optional): The file format of the :py:attr:`trajectory`, for a trajectory file path to be read by :py:class:`pymatgen.io.vasp.Xdatcar` this should be :py:attr:`'Xdatcar'`, for a :py:attr:`list` of :py:class:`pymatgen.core.structure.Structure` objects this should be :py:attr:`'structures'`, for a trajectory file path to be read by :py:mod:`MDAnalysis` this should be the appropriate format to be passed to the :py:class:`MDAnalysis.core.universe.Universe`, and for a n :py:class:`MDAnalysis.core.universe.Universe` object this should be :py:attr:`'universe'`. Defaults to :py:attr:`'Xdatcar'`.
        n_samples_fit (:py:attr:`int`, optional): The number of samples in the random generator for the fitting of the linear relationship. Default is :py:attr:`10000`.
        fit_intercept (:py:attr:`bool`, optional): Should the intercept of the diffusion relationship be fit. Default is :py:attr:`True`.
    """
    def __init__(self, trajectory, parser_params, bootstrap_params=None, dtype='Xdatcar', n_samples_fit=10000, fit_intercept=True):  # pragma: no cover
        super().__init__(trajectory, parser_params, bootstrap_params, dtype, charge)
        self._diff.diffusion(n_samples_fit, fit_intercept)

    @property
    def D(self):
        """
        Diffusion coefficient.

        Returns:
            :py:class:`uravu.distribution.Distribution`: Diffusion coefficient.
        """
        return self._diff.diffusion_coefficient

    @property
    def D_offset(self):
        """
        Offset from abscissa.

        Returns:
            :py:class:`uravu.distribution.Distribution`: Abscissa offset.
        """
        return self._diff.intercept


def _flatten_list(this_list):
    """
    Flatten nested lists.

    Args:
        this_list (:py:attr:`list`): List to be flattened.

    Returns:
        :py:attr:`list`: Flattened list.
    """
    return [item for sublist in this_list for item in sublist]



"""
Tests for diffusion module

Copyright (c) Andrew R. McCluskey and Benjamin J. Morgan

Distributed under the terms of the MIT License

@author: Andrew R. McCluskey
"""

# pylint: disable=R0201

import unittest
import numpy as np
import uncertainties
from numpy.testing import assert_equal, assert_almost_equal
from scipy.stats import norm
from kinisi import diffusion
from uravu.distribution import Distribution
from uravu.utils import straight_line


dt = np.linspace(5, 50, 10)
msd = np.linspace(5, 50, 10)
MSD = []
for i in msd:
    MSD.append(Distribution(norm.rvs(loc=i, scale=i*0.1, size=5000, random_state=np.random.RandomState(1))))

class TestMsd(unittest.TestCase):
    """
    Unit tests for diffusion module
    """

    def test_msd_bootstrap_a(self):
        """
        Test msd_bootstrap for initial normal.
        """
        ordinate = np.random.randn(100, 50, 3)
        to_resample = [
            np.array(ordinate) for i in range(1, 6)
        ]
        boot = diffusion.MSDBootstrap(
            np.linspace(100, 600, 5, dtype=int),
            to_resample,
            progress=False,
            n_resamples=100,
            max_resamples=1000)
        assert_equal(boot.dt.size, 5)
        assert_equal(len(boot.distributions), 5)
        assert_equal(boot.msd_observed.size, 5)
        assert_equal(boot.msd_sampled.size, 5)

    def test_msd_bootstrap_b(self):
        """
        Test msd_bootstrap for initial uniform with progress.
        """
        ordinate = np.random.uniform(size=(100, 50, 3))
        to_resample = [
            np.array(ordinate) for i in range(1, 6)
        ]
        boot = diffusion.MSDBootstrap(
            np.linspace(100, 600, 5, dtype=int),
            to_resample,
            progress=True,
            n_resamples=1,
            max_resamples=1000)
        assert_equal(boot.dt.size, 5)
        assert_equal(len(boot.distributions), 5)
        assert_equal(boot.msd_observed.size, 5)
        assert_equal(boot.msd_sampled.size, 5)

    def test_msd_bootstrap_c(self):
        """
        Test msd_bootstrap for to go over the limit point.
        """
        ordinate1 = np.random.randn(5, 500, 3)
        ordinate2 = np.random.randn(5, 450, 3)
        ordinate3 = np.random.randn(5, 400, 3)
        ordinate4 = np.random.randn(5, 100, 3)
        ordinate5 = np.random.randn(5, 10, 3)
        to_resample = [
            ordinate1,
            ordinate2,
            ordinate3,
            ordinate4,
            ordinate5,
        ]
        boot = diffusion.MSDBootstrap(
            np.linspace(100, 600, 5, dtype=int),
            to_resample,
            progress=False,
            n_resamples=1,
            max_resamples=10,
            sub_sample_dt=2)
        assert_equal(boot.dt.size, 3)
        assert_equal(len(boot.distributions), 3)
        assert_equal(boot.msd_observed.size, 3)
        assert_equal(boot.msd_sampled.size, 3)

    def test_msd_bootstrap_d(self):
        """
        Test msd_bootstrap very few particles.
        """
        ordinate1 = np.random.randn(10, 10, 3)
        ordinate2 = np.random.randn(1, 1, 3)
        ordinate3 = np.random.randn(1, 1, 3)
        ordinate4 = np.random.randn(1, 1, 3)
        ordinate5 = np.random.randn(1, 1, 3)
        to_resample = [
            ordinate1,
            ordinate2,
            ordinate3,
            ordinate4,
            ordinate5,
        ]
        boot = diffusion.MSDBootstrap(
            np.linspace(100, 600, 5, dtype=int),
            to_resample,
            progress=False,
            n_resamples=1,
            max_resamples=10,
            sub_sample_dt=2)
        assert_equal(boot.dt.size, 1)
        assert_equal(len(boot.distributions), 1)
        assert_equal(boot.msd_observed.size, 1)
        assert_equal(boot.msd_sampled.size, 1)



"""
The two classes herein enable the determination of the activation energy from a diffusion-Arrhenius or -Super-Arrhenius plot.
This includes the uncertainty on the activation energy from the MCMC sampling of the plot, with uncertainties on diffusion.
Furthermore, the classes are build on the Relationship subclass, therefore it is easy to determine the Bayesian evidence for each of the models with the given data, enabling the distinction between Arrhenius and Super-Arrhenius diffusion.
"""

# Copyright (c) Andrew R. McCluskey and Benjamin J. Morgan
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

# pylint: disable=R0913

import sys
import numpy as np
from uravu.relationship import Relationship
from scipy.constants import R, N_A, eV
from scipy.stats import uniform

# Convert R to eV
R_no_mol = R / N_A
R_eV = R_no_mol / eV


class StandardArrhenius(Relationship):
    """
    Evaluate the data with a standard Arrhenius relationship.
    For attributes associated with the :py:class:`uravu.relationship.Relationship` class see that documentation.
    The :py:attr:`uravu.relationship.Relationship.variables` for this model is a :py:attr:`list` of length 2, where :py:attr:`kinisi.arrhenius.StandardArrhenius.variables[0]` is the activation energy (in eV) and :py:attr:`kinisi.arrhenius.StandardArrhenius.variables[1]` is the prefactor of the Arrhenius equation.

    Args:
        temperature (:py:attr:`array_like`): Temperature data.
        diffusion (:py:attr:`array_like`): Diffusion coefficient data.
        bounds (:py:attr:`tuple`): The minimum and maximum values for each parameters. Defaults to :py:attr:`None`.
        diffusion_error (:py:attr:`array_like`): Uncertainty in the diffusion coefficient data. Not necessary if :py:attr:`diffusion` is :py:attr:`list` of :py:class:`uravu.distribution.Distribution` objects.
        ci_points (:py:attr:`array_like`, optional): The two percentiles at which confidence intervals should be found for the variables. Default is :py:attr:`[2.5, 97.5]` (a 95 % confidence interval).
    """
    def __init__(self, temperature, diffusion, bounds, diffusion_error=None, ci_points=None):
        super().__init__(arrhenius, temperature, diffusion, bounds, ordinate_error=diffusion_error, ci_points=None)


def arrhenius(abscissa, activation_energy, prefactor):
    """
    Determine the diffusion coefficient for a given activation energy, and prefactor according to the Arrhenius equation.

    Args:
        abscissa (:py:attr:`array_like`): The abscissa data.
        activation_energy (:py:attr:`float`): The activation_energy value.
        prefactor (:py:attr:`float`): The prefactor value.

    Returns:
        :py:attr:`array_like`: The diffusion coefficient data.
    """
    return prefactor * np.exp(-1 * activation_energy / (R_eV * abscissa))


class SuperArrhenius(Relationship):
    """
    Evaluate the data with a super-Arrhenius relationship. For attributes associated with the :py:class:`uravu.relationship.Relationship` class see that documentation.
    This :py:attr:`uravu.relationship.Relationship.variables` for this model is a :py:attr:`list` of length 3, where :py:attr:`~uravu.relationship.Relationship.variables[0]` is the activation energy (in eV), :py:attr:`~uravu.relationship.Relationship.variables[1]` is the prefactor of the Arrhenius equation, and :py:attr:`~uravu.relationship.Relationship.variables[2]` is the temperature offset (in Kelvin).

    Args:
        temperature (:py:attr:`array_like`): Temperature data.
        diffusion (:py:attr:`array_like`): Diffusion coefficient data.
        bounds (:py:attr:`tuple`): The minimum and maximum values for each parameters. Defaults to :py:attr:`None`.
        diffusion_error (:py:attr:`array_like`): Uncertainty in the diffusion coefficient data. Not necessary if :py:attr:`diffusion` is :py:attr:`list` of :py:class:`uravu.distribution.Distribution` objects.
        ci_points (:py:attr:`array_like`, optional): The two percentiles at which confidence intervals should be found for the variables. Default is :py:attr:`[2.5, 97.5]` (a 95 % confidence interval).
    """
    def __init__(self, temperature, diffusion, bounds, diffusion_error=None, ci_points=None):
        super().__init__(super_arrhenius, temperature, diffusion, bounds, ordinate_error=diffusion_error, ci_points=None)


def super_arrhenius(abscissa, activation_energy, prefactor, t_zero):
    """
    Determine the rate constant for a given activation energy, prefactor, and t_zero according to the Vogel–Tammann–Fulcher equation.

    Args:
        abscissa (:py:attr:`array_like`): The abscissa data.
        activation_energy (:py:attr:`float`): The activation_energy value.
        prefactor (:py:attr:`float`): The prefactor value.
        t_zero (:py:attr:`float`): The T_0 value.

    Returns:
        :py:attr:`array_like`: The diffusion coefficient data.
    """
    return prefactor * np.exp(
        -1 * activation_energy / (R_eV * (abscissa - t_zero)))



"""
The modules is focused on tools for the evaluation of the mean squared displacement and resulting diffusion coefficient from a material.
"""

# Copyright (c) Andrew R. McCluskey and Benjamin J. Morgan
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

# pylint: disable=R0913,R0914

import sys
import warnings
import numpy as np
import pandas as pd
from scipy.stats import uniform, norm, multivariate_normal
from sklearn.utils import resample
from tqdm import tqdm
from statsmodels.stats.moment_helpers import corr2cov
from uravu.distribution import Distribution
from uravu.axis import Axis


class Bootstrap:
    """
    The top-level class for bootstrapping.

    Attributes:
        confidence_interval (:py:attr:`array_like`): The percentile points of the distribution that should be stored.
        displacements (:py:attr:`list` of :py:attr:`array_like`): A list of arrays, where each array has the axes [atom, displacement observation, dimension]. There is one array in the list for each delta_t value. Note: it is necessary to use a list of arrays as the number of observations is not necessary the same at each data point.
        delta_t (:py:attr:`array_like`): An array of the timestep values.
        max_obs (:py:attr:`int`): The maximum number of observations for the trajectory.
        dt (:py:attr:`array_like`): Timestep values that were resampled.
        distributions ():py:attr:`list` of :py:class:`uravu.distribution.Distribution`): Resampled mean squared distributions.
        iterator (:py:attr:`tqdm` or :py:attr:`range`): The iteration object.


    Args:
        delta_t (:py:attr:`array_like`): An array of the timestep values.
        disp_3d (:py:attr:`list` of :py:attr:`array_like`): A list of arrays, where each array has the axes [atom, displacement observation, dimension]. There is one array in the list for each delta_t value. Note: it is necessary to use a list of arrays as the number of observations is not necessary the same at each data point.
        sub_sample_dt (:py:attr:`int`. optional): The frequency in observations to be sampled. Default is :py:attr:`1` (every observation).
        confidence_interval (:py:attr:`array_like`, optional): The percentile points of the distribution that should be stored. Default is :py:attr:`[2.5, 97.5]` (a 95 % confidence interval).
        progress (:py:attr:`bool`, optional): Show tqdm progress for sampling. Default is :py:attr:`True`.
    """
    def __init__(self, delta_t, disp_3d, sub_sample_dt=1, confidence_interval=None, progress=True):
        if confidence_interval is None:
            self.confidence_interval = [2.5, 97.5]
        self.displacements = disp_3d[::sub_sample_dt]
        self.delta_t = np.array(delta_t[::sub_sample_dt])
        self.max_obs = self.displacements[0].shape[1]
        self.distributions = []
        self.dt = np.array([])
        self.iterator = _iterator(progress, range(len(self.displacements)))
        self.diffusion_coefficient = None
        self.intercept = None

    @property
    def D(self):
        """
        An alias for the diffusion coefficient Distribution.
        """
        return self.diffusion_coefficient


class MSDBootstrap(Bootstrap):
    """
    Perform a bootstrap resampling to obtain accurate estimates for the mean and uncertainty for the squared displacements.
    This resampling method is applied until the MSD distribution is normal (or the `max_resamples` has been reached) and therefore may be described with a median and confidence interval.

    Attributes:
        msd_observed (:py:attr:`array_like`): The sample mean-squared displacements, found from the arithmetic average of the observations.
        msd_sampled (:py:attr:`array_like`): The population mean-squared displacements, found from the bootstrap resampling of the observations.
        msd_sampled_err (:py:attr:`array_like`): The two-dimensional uncertainties, at the given confidence interval, found from the bootstrap resampling of the observations.
        correlation_matrix (:py:attr:`array_like`): The estimated correlation matrix for the mean-squared displacements.

    Args:
        delta_t (:py:attr:`array_like`): An array of the timestep values.
        disp_3d (:py:attr:`list` of :py:attr:`array_like`): A list of arrays, where each array has the axes [atom, displacement observation, dimension]. There is one array in the list for each delta_t value. Note: it is necessary to use a list of arrays as the number of observations is not necessary the same at each data point.
        n_resamples (:py:attr:`int`, optional): The initial number of resamples to be performed. Default is :py:attr:`1000`.
        sub_sample_dt (:py:attr:`int`. optional): The frequency in observations to be sampled. Default is :py:attr:`1` (every observation).
        confidence_interval (:py:attr:`array_like`, optional): The percentile points of the distribution that should be stored. Default is :py:attr:`[31.73, 68.27]` (a single standard deviation).
        max_resamples (:py:attr:`int`, optional): The max number of resamples to be performed by the distribution is assumed to be normal. This is present to allow user control over the time taken for the resampling to occur. Default is :py:attr:`100000`.
        bootstrap_multiplier (:py:attr:`int`, optional): The factor by which the number of bootstrap samples should be multiplied. The default is :py:attr:`1`, which is the maximum number of truely independent samples in a given timestep. This can be increase, however it is importance to note that when greater than 1 the sampling is no longer independent.
        progress (:py:attr:`bool`, optional): Show tqdm progress for sampling. Default is :py:attr:`True`.
    """
    def __init__(self, delta_t, disp_3d, n_resamples=1000, sub_sample_dt=1, confidence_interval=None, max_resamples=10000, bootstrap_multiplier=1, progress=True):
        super().__init__(delta_t, disp_3d, sub_sample_dt, confidence_interval, progress)
        self.msd_observed = np.array([])
        self.msd_sampled = np.array([])
        self.msd_sampled_err = np.array([])
        samples = np.zeros((self.displacements[0].shape[0], len(self.displacements)))
        for i in self.iterator:
            d_squared = np.sum(self.displacements[i] ** 2, axis=2)
            samples[:, i] = d_squared.mean(axis=1).flatten()
            n_samples_msd = _n_samples(self.displacements[i].shape, self.max_obs, bootstrap_multiplier)
            if n_samples_msd <= 1:
                continue
            self.msd_observed = np.append(self.msd_observed, np.mean(d_squared.flatten()))
            distro = _sample_until_normal(d_squared, n_samples_msd, n_resamples, max_resamples, self.confidence_interval)
            self.dt = np.append(self.dt, self.delta_t[i])
            self.distributions.append(distro)
            self.msd_sampled = np.append(self.msd_sampled, distro.n)
            self.msd_sampled_err = np.append(self.msd_sampled_err, np.std(distro.samples))
        self.correlation_matrix = np.array(pd.DataFrame(samples).corr())

    def diffusion(self, n_samples=10000, fit_intercept=True):
        """
        Calculate the diffusion coefficient for the trajectory.

        Args:
            n_samples (:py:attr:`int`, optional): The number of samples in the random generator. Default is :py:attr:`10000`.
            fit_intercept (:py:attr:`bool`, optional): Should the intercept of the diffusion relationship be fit. Default is :py:attr:`True`.
        """
        single_msd = multivariate_normal(self.msd_sampled, corr2cov(self.correlation_matrix, self.msd_sampled_err), allow_singular=True)
        single_msd_samples = single_msd.rvs(n_samples)
        A = np.array([self.dt]).T
        if fit_intercept:
            A = np.array([np.ones(self.dt.size), self.dt]).T
        Y = single_msd_samples.T
        straight_line = np.matmul(np.matmul(np.linalg.inv(np.matmul(A.T, A)), A.T), Y)
        if fit_intercept:
            intercept, gradient = straight_line
            self.diffusion_coefficient = Distribution(gradient / 6, ci_points=self.confidence_interval)
            self.intercept = Distribution(intercept, ci_points=self.confidence_interval)
        else:
            self.diffusion_coefficient = Distribution(straight_line[0] / 6, ci_points=self.confidence_interval)


def _n_samples(disp_shape, max_obs, bootstrap_multiplier):
    """
    Calculate the maximum number of independent observations.

    Args:
        disp_shape (:py:attr:`tuple`): The shape of the displacements array.
        max_obs (:py:attr:`int`): The maximum number of observations for the trajectory.
        bootstrap_multiplier (:py:attr:`int`, optional): The factor by which the number of bootstrap samples should be multiplied. The default is :py:attr:`1`, which is the maximum number of truely independent samples in a given timestep. This can be increase, however it is importance to note that when greater than 1 the sampling is no longer independent.

    Returns:
        :py:attr:`int`: Maximum number of independent observations.
    """
    n_obs = disp_shape[1]
    n_atoms = disp_shape[0]
    dt_int = max_obs - n_obs + 1
    # approximate number of "non-overlapping" observations, allowing
    # for partial overlap
    return int(max_obs / dt_int * n_atoms) * bootstrap_multiplier


def _iterator(progress, loop):
    """
    Get the iteration object, using :py:mod:`tqdm` as appropriate.

    Args:
        progress (:py:attr:`bool`): Should :py:mod:`tqdm` be used to give a progress bar.
        loop (:py:attr:`list` or :py:attr:`range`): The object that should be looped over.

    Returns:
        :py:attr:`tqdm` or :py:attr:`range`: Iterator object.
    """
    if progress:
        return tqdm(loop, desc='Bootstrapping Displacements')
    else:
        return loop


def _sample_until_normal(array, n_samples, n_resamples, max_resamples, confidence_interval):
    """
    Resample from the distribution until a normal distribution is obtained or a maximum is reached.

    Args:
        array (:py:attr:`array_like`): The array to sample from.
        n_samples (:py:attr:`int`): Number of samples.
        r_resamples (:py:attr:`int`): Number of resamples to perform initially.
        max_resamples (:py:attr:`int`): The maximum number of resamples to perform.
        confidence_interval (:py:attr:`array_like`): The percentile points of the distribution that should be stored.

    Returns:
        :py:class:`uravu.distribution.Distribution`: The resampled distribution.
    """
    distro = Distribution(_bootstrap(array.flatten(), n_samples, n_resamples), ci_points=confidence_interval)
    while (not distro.normal) and distro.size < max_resamples:
        distro.add_samples(_bootstrap(array.flatten(), n_samples, 100))
    if distro.size >= max_resamples:
        warnings.warn("The maximum number of resamples has been reached, and the distribution is not yet normal.")
    return distro


def _bootstrap(array, n_samples, n_resamples):
    """
    Perform a set of resamples.

    Args:
        array (:py:attr:`array_like`): The array to sample from.
        n_samples (:py:attr:`int`): Number of samples.
        r_resamples (:py:attr:`int`): Number of resamples to perform initially.

    Returns:
        :py:attr:`array_like`: Resampled values from the array.
    """
    return [np.mean(resample(array.flatten(), n_samples=n_samples)) for j in range(n_resamples)]



"""
Parser functions, including implementation for :py:mod:`pymatgen` compatible VASP files and :py:mod:`MDAnalysis` compatible trajectories.
"""

# Copyright (c) Andrew R. McCluskey and Benjamin J. Morgan
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

# pylint: disable=R0902,R0913

# This parser borrows heavily from the pymatgen.analysis.diffusion_analyzer.DiffusionAnalyzer class, originally authored by Will Richards (wrichard@mit.edu) and Shyue Ping Ong. We include this statement to not that we make no claim to authorship of that code and make no attack on the original authors.

# In fact, we love pymatgen!

import numpy as np
import periodictable as pt
from tqdm import tqdm


class Parser:
    """
    The base class for parsing.

    Attributes:
        time_step (:py:attr:`float`): Time step between measurements.
        step_step (:py:attr:`int`): Sampling freqency of the displacements (time_step is multiplied by this number to get the real time between measurements).
        indices (:py:attr:`array_like`): Indices for the atoms in the trajectory used in the calculation of the diffusion.
        delta_t (:py:attr:`array_like`):  Timestep values.
        disp_3d (:py:attr:`list` of :py:attr:`array_like`): Each element in the :py:attr:`list` has the axes [atom, displacement observation, dimension] and there is one element for each delta_t value. *Note: it is necessary to use a :py:attr:`list` of :py:attr:`array_like` as the number of observations is not necessary the same at each timestep point*.
        min_dt (:py:attr:`float`, optional): Minimum timestep to be evaluated.

    Args:
        disp (:py:attr:`array_like`): Displacements of atoms.
        structures (:py:attr:`list` of :py:class`pymatgen.core.structure.Structure`): Structures ordered in sequence of run.
        specie (:py:class:`pymatgen.core.periodic_table.Element` or :py:class:`pymatgen.core.periodic_table.Specie`): Specie to calculate diffusivity for as a String, e.g. :py:attr:`'Li'`.
        time_step (:py:attr:`float`): Time step between measurements.
        step_step (:py:attr:`int`): Sampling freqency of the displacements (time_step is multiplied by this number to get the real time between measurements).
        min_obs (:py:attr:`int`, optional): Minimum number of observations to have before including in the MSD vs dt calculation. E.g. If a structure has 10 diffusing atoms, and :py:attr:`min_obs=30`, the MSD vs dt will be calculated up to :py:attr:`dt = total_run_time / 3`, so that each diffusing atom is measured at least 3 uncorrelated times. Defaults to :py:attr:`30`.
        min_dt (:py:attr:`float`, optional): Minimum timestep to be evaluated. Defaults to :py:attr:`100`.
        ndt (:py:attr:`int`, optional): The number of dt values to calculate the MSD over. Defaults to :py:attr:`75`.
        progress (:py:attr:`bool`, optional): Print progress bars to screen. Defaults to :py:attr:`True`.
    """
    def __init__(self, disp, indices, framework_indices, time_step, step_skip, min_obs=30, min_dt=100, ndt=75, progress=True):
        self.time_step = time_step
        self.step_skip = step_skip
        self.indices = indices
        self.min_dt = min_dt
        self.ndt = ndt

        drift_corrected = self.correct_drift(framework_indices, disp)

        nsteps = drift_corrected.shape[1]

        timesteps = self.smoothed_timesteps(nsteps, min_obs, indices)

        self.delta_t, self.disp_3d = self.get_disps(
            timesteps, drift_corrected, progress)

    def smoothed_timesteps(self, nsteps, min_obs, indices):
        """
        Calculate the smoothed timesteps to be used.

        Args:
            nsteps (:py:attr:`int`): Number of time steps.
            min_obs (:py:attr:`int`): Minimum number of observations to have before including in the MSD vs dt calculation. E.g. If a structure has 10 diffusing atoms, and :py:attr:`min_obs=30`, the MSD vs dt will be calculated up to :py:attr:`dt = total_run_time / 3`, so that each diffusing atom is measured at least 3 uncorrelated times.
            indices (:py:attr:`array_like`): Indices for the atoms in the trajectory used in the calculation of the diffusion.

        Returns:
            :py:attr:`array_like`: Smoothed timesteps.
        """
        min_dt = int(self.min_dt / (self.step_skip * self.time_step))
        max_dt = min(len(indices) * nsteps // min_obs, nsteps)
        if min_dt == 0:
            min_dt = 1
        if min_dt >= max_dt:
            raise ValueError('Not enough data to calculate diffusivity')
        timesteps = np.arange(min_dt, max_dt, max(int((max_dt - min_dt) / self.ndt), 1))
        return timesteps

    def get_disps(self, timesteps, drift_corrected, progress=True):
        """
        Calculate the mean-squared displacement

        Args:
            timesteps (:py:attr:`array_like`): Smoothed timesteps.
            drift_corrected (:py:attr:`array_like`): Drift of framework corrected disp.
            progress (:py:attr:`bool`, optional): Print progress bars to screen. Defaults to :py:attr:`True`.

        Returns:
            :py:attr:`tuple`: Containing:
                - :py:attr:`array_like`: Time step intervals.
                - :py:attr:`array_like`: Raw squared displacement.
        """
        delta_t = timesteps * self.time_step * self.step_skip
        disp_3d = []
        if progress:
            iterator = tqdm(timesteps, desc='Getting Displacements')
        else:
            iterator = timesteps
        for timestep in iterator:
            disp = np.subtract(drift_corrected[self.indices, timestep:, :], drift_corrected[self.indices, :-timestep, :])
            disp_3d.append(disp)
        return delta_t, disp_3d

    def correct_drift(self, framework_indices, disp):
        """
        Perform drift correction

        Args:
            framework_indices (:py:attr:`array_like`): Indices of framework atoms, to correct against.
            disp (:py:attr:`array_like`): Numpy array of with shape [site, time step, axis].

        Returns:
            :py:attr:`array_like`: Drift of framework corrected disp.
        """
        # drift corrected position
        if len(framework_indices) > 0:
            framework_disp = disp[framework_indices]
            drift_corrected = disp - np.average(framework_disp, axis=0)[None, :, :]
        else:
            drift_corrected = disp

        return drift_corrected


class PymatgenParser(Parser):
    """
    A parser for pymatgen structures.

    Args:
        structures (:py:attr:`list` or :py:class`pymatgen.core.structure.Structure`): Structures ordered in sequence of run.
        specie (:py:class:`pymatgen.core.periodic_table.Element` or :py:class:`pymatgen.core.periodic_table.Specie`): Specie to calculate diffusivity for as a String, e.g. :py:attr:`'Li'`.
        time_step (:py:attr:`float`): Time step between measurements.
        step_step (:py:attr:`int`): Sampling frequency of the displacements (time_step is multiplied by this number to get the real time between measurements).
        sub_sample_traj (:py:attr:`float`, optional): Multiple of the :py:attr:`time_step` to sub sample at. Defaults to :py:attr:`1` where all timesteps are used.
        min_obs (:py:attr:`int`, optional): Minimum number of observations to have before including in the MSD vs dt calculation. E.g. If a structure has 10 diffusing atoms, and :py:attr:`min_obs=30`, the MSD vs dt will be calculated up to :py:attr:`dt = total_run_time / 3`, so that each diffusing atom is measured at least 3 uncorrelated times. Defaults to :py:attr:`30`.
        min_dt (:py:attr:`float`, optional): Minimum timestep to be evaluated. Defaults to :py:attr:`100`.
        ndt (:py:attr:`int`, optional): The number of dt values to calculate the MSD over. Defaults to :py:attr:`75`.
        progress (:py:attr:`bool`, optional): Print progress bars to screen. Defaults to :py:attr:`True`.
    """
    def __init__(self, structures, specie, time_step, step_skip, min_obs=30, sub_sample_traj=1, min_dt=100, ndt=75, progress=True):
        structure, coords, latt = _pmg_get_structure_coords_latt(structures, sub_sample_traj, progress)

        disp, latt = _get_disp(coords, latt)

        indices, framework_indices = self.get_indices(structure, specie)

        super().__init__(disp, indices, framework_indices, time_step, step_skip, min_obs, min_dt, ndt, progress)

    def get_indices(self, structure, specie):
        """
        Determine framework and non-framework indices

        Perform drift correction

        Args:
            structure (:py:class:`pymatgen.core.structure.Structure`): Initial structure.
            specie (:py:class:`pymatgen.core.periodic_table.Element` or :py:class:`pymatgen.core.periodic_table.Specie`): Specie to calculate diffusivity for as a String, e.g. :py:attr:`'Li'`.

        Returns:
            :py:attr:`array_like`: Indices for the atoms in the trajectory used in the calculation of the diffusion.
            :py:attr:`array_like`: Indices of framework atoms.
        """
        indices = []
        framework_indices = []
        for i, site in enumerate(structure):
            if site.specie.symbol == specie:
                indices.append(i)
            else:
                framework_indices.append(i)
        return indices, framework_indices


class MDAnalysisParser(Parser):
    """
    A parser that consumes an MDAnalysis.Universe object.

    Args:
        universe (:py:class:`MDAnalysis.core.universe.Universe`): The MDAnalysis object of interest.
        specie (:py:attr:`str`): Specie to calculate diffusivity for as a String, e.g. :py:attr:`'Li'`.
        time_step (:py:attr:`float`): Time step between measurements.
        step_step (:py:attr:`int`): Sampling freqency of the displacements (time_step is multiplied by this number to get the real time between measurements).
        sub_sample_traj (:py:attr:`float`, optional): Multiple of the :py:attr:`time_step` to sub sample at. Defaults to :py:attr:`1` where all timesteps are used.
        min_obs (:py:attr:`int`, optional): Minimum number of observations to have before including in the MSD vs dt calculation. E.g. If a structure has 10 diffusing atoms, and :py:attr:`min_obs=30`, the MSD vs dt will be calculated up to :py:attr:`dt = total_run_time / 3`, so that each diffusing atom is measured at least 3 uncorrelated times. Defaults to :py:attr:`30`.
        min_dt (:py:attr:`float`, optional): Minimum timestep to be evaluated. Defaults to :py:attr:`100`.
        ndt (:py:attr:`int`, optional): The number of dt values to calculate the MSD over. Defaults to :py:attr:`75`.
        progress (:py:attr:`bool`, optional): Print progress bars to screen. Defaults to :py:attr:`True`.
    """
    def __init__(self, universe, specie, time_step, step_skip, min_obs=30, sub_sample_traj=1, min_dt=100, ndt=75, progress=True):
        structure, coords, latt = _mda_get_structure_coords_latt(universe, specie, sub_sample_traj, progress)

        disp, latt = _get_disp(coords, latt)

        indices, framework_indices = self.get_indices(structure, specie)

        super().__init__(disp, indices, framework_indices, time_step, step_skip * sub_sample_traj, min_obs, min_dt, ndt, progress)

    def get_indices(self, structure, specie):
        """
        Determine framework and non-framework indices

        Perform drift correction

        Args:
            structure (:py:class:`MDAnalysis.core.groups.AtomGroup`): Initial structure.
            specie (:py:class:`pymatgen.core.periodic_table.Element` or :py:class:`pymatgen.core.periodic_table.Specie`): Specie to calculate diffusivity for as a String, e.g. :py:attr:`'Li'`.

        Returns:
            :py:attr:`array_like`: Indices for the atoms in the trajectory used in the calculation of the diffusion.
            :py:attr:`array_like`: Indices of framework atoms.
        """
        indices = []
        framework_indices = []
        species_names = []
        if not isinstance(specie, list):
            specie = [specie]
        for i, site in enumerate(structure):
            if site.type in specie:
                indices.append(i)
            else:
                framework_indices.append(i)
        return indices, framework_indices


def _mda_get_structure_coords_latt(universe, specie, sub_sample_traj=1, progress=True):
    """
    Obtain the initial structure and displacement from a :py:class:`MDAnalysis.universe.Universe` file

    Args:
        universe (:py:class:MDAnalysis.universe.Universe): Universe for analysis.
        specie (:py:attr:`str`): Specie to calculate diffusivity for as a String, e.g. :py:attr:`'Li'`.
        sub_sample_traj (:py:attr:`float`, optional): Multiple of the :py:attr:`time_step` to sub sample at. Default is :py:attr:`1`.
        progress (:py:attr:`bool`, optional): Print progress bars to screen. Defaults to :py:attr:`True`.

    Returns:
        :py:class`pymatgen.core.structure.Structure`: Initial structure.
        :py:attr:`list` of :py:attr:`array_like`: Fractional coordinates for all atoms.
        :py:attr:`list` of :py:attr:`array_like`: Lattice descriptions.
    """
    coords, latt = [], []
    first = True
    if progress:
        iterator = tqdm(universe.trajectory[::sub_sample_traj], desc='Reading Trajectory')
    else:
        iterator = universe.trajectory[::sub_sample_traj]
    for timestep in iterator:
        if first:
            structure = universe.atoms
            first = False
        matrix = get_matrix(timestep.dimensions)
        inv_matrix = np.linalg.inv(matrix)
        coords.append(np.array(np.dot(universe.atoms.positions, inv_matrix))[:, None])
        latt.append(matrix)
    coords.insert(0, coords[0])
    latt.insert(0, latt[0])
    return structure, coords, latt


def get_matrix(dimensions):
    """
    Determine the lattice matrix.

    Args:
        dimensions (:py:attr:`array_like`): a, b, c, vectors and alpha, beta, gamma angles.

    Returns:
        :py:attr:`array_like`: Lattice matrix
    """
    a, b, c, alpha, beta, gamma = dimensions

    angles_r = np.radians([alpha, beta, gamma])
    cos_alpha, cos_beta, cos_gamma = np.cos(angles_r)
    sin_alpha, sin_beta, sin_gamma = np.sin(angles_r)

    val = (cos_alpha * cos_beta - cos_gamma) / (sin_alpha * sin_beta)
    # Sometimes rounding errors result in values slightly > 1.
    val = max(min(val, 1), -1)
    gamma_star = np.arccos(val)

    vector_a = [a * sin_beta, 0.0, a * cos_beta]
    vector_b = [-b * sin_alpha * np.cos(gamma_star), b * sin_alpha * np.sin(gamma_star), b * cos_alpha]
    vector_c = [0.0, 0.0, float(c)]

    return np.array([vector_a, vector_b, vector_c], dtype=np.float64).reshape((3, 3))


def _pmg_get_structure_coords_latt(structures, sub_sample_traj=1, progress=True):
    """
    Obtain the initial structure and displacement from a :py:attr:`list` of :py:class`pymatgen.core.structure.Structure`.

    Args:
        structures (:py:attr:`list` of :py:class`pymatgen.core.structure.Structure`): Structures ordered in sequence of run.
        sub_sample_traj (:py:attr:`float`, optional): Multiple of the :py:attr:`time_step` to sub sample at. Default is :py:attr:`1`.
        progress (:py:attr:`bool`, optional): Print progress bars to screen. Defaults to :py:attr:`True`.

    Returns:
        :py:class`pymatgen.core.structure.Structure`: Initial structure.
        :py:attr:`list` of :py:attr:`array_like`: Fractional coordinates for all atoms.
        :py:attr:`list` of :py:attr:`array_like`: Lattice descriptions.
    """
    coords, latt = [], []
    first = True
    if progress:
        iterator = tqdm(structures[::sub_sample_traj], desc='Reading Trajectory')
    else:
        iterator = structures[::sub_sample_traj]
    for struct in iterator:
        if first:
            structure = struct
            first = False
        coords.append(np.array(struct.frac_coords)[:, None])
        latt.append(struct.lattice.matrix)
    coords.insert(0, coords[0])
    latt.insert(0, latt[0])

    return structure, coords, latt

def _get_disp(coords, latt):
    """
    Calculate displacements.

    Args:
        coords (:py:attr:`list` of :py:attr:`array_like`): Fractional coordinates for all atoms.
        latt (:py:attr:`list` of :py:attr:`array_like`): Lattice descriptions.

    Returns:
        :py:attr:`array_like`: Numpy array of with shape [site, time step, axis] describing displacements.
    """
    coords = np.concatenate(coords, axis=1)
    d_coords = coords[:, 1:] - coords[:, :-1]
    d_coords = d_coords - np.round(d_coords)
    f_disp = np.cumsum(d_coords, axis=1)
    c_disp = []
    for i in f_disp:
        c_disp.append([np.dot(d, m) for d, m in zip(i, latt[1:])])
    disp = np.array(c_disp)

    # If is NVT-AIMD, clear lattice data.
    if np.array_equal(latt[0], latt[-1]):
        latt = np.array([latt[0]])
    else:
        latt = np.array(latt)

    return disp, latt
